// (c) 2006-2010 Philipp Klaus Krause philipp@colecovision.eu

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

//#include "rle_escape.h"
#define RLE_ESCAPE 0

#define BUFFER_SIZE 128

struct ringbuffer
{
	unsigned char buffer[BUFFER_SIZE];
	unsigned char start;
};

void ringbuffer_insert(struct ringbuffer *buffer, const unsigned char c)
{
	buffer->buffer[++(buffer->start)] = c;
	buffer->start %= BUFFER_SIZE;
}

unsigned char ringbuffer_get(const struct ringbuffer *buffer, const unsigned char offset)
{
	return(buffer->buffer[(buffer->start + BUFFER_SIZE - offset) % BUFFER_SIZE]);
}

unsigned int offsets[128];

unsigned char get_seq_length(FILE *in, const struct ringbuffer *buffer, const unsigned char offset)
{
	unsigned char length;
	fpos_t pos;
	struct ringbuffer buffer2 = *buffer;

	fgetpos(in, &pos);
	for(length = 0; length <= 33; length++)
	{
		int fc = getc(in);
		if(fc == EOF || ringbuffer_get(&buffer2, offset) != fc)
			break;
		ringbuffer_insert(&buffer2, fc);
	}
	fsetpos(in, &pos);

	return(length < 33 ? length : 33);
}

void stream_to_buffer(struct ringbuffer *buffer, FILE *in, unsigned char length)
{
	unsigned char fc, i;
	for(i = 0; i < length; i++)
	{
		fc = getc(in);
		ringbuffer_insert(buffer, fc);
	}
}

void lzk_encode(FILE *in, FILE *out)
{
	for(int z = 0; z < 128; z++) offsets[z] = 0;

	unsigned char offset, bestoffset, bestlength, tmplength;
	struct ringbuffer buffer;

	memset(buffer.buffer, 0xff, BUFFER_SIZE);
	buffer.start = 0;

	int fc;
	for(;;)
	{
		fc = getc(in);
		if(fc == EOF)
			break;
		ungetc(fc, in);

		for(offset = 0, bestoffset = 0, bestlength = 0; offset < BUFFER_SIZE; offset++)
		{
			if((tmplength = get_seq_length(in, &buffer, offset)) <= bestlength)
				continue;

			bestlength = tmplength;
			bestoffset = offset;
		}
if(bestlength > 3)
printf("bo: %d, bl: %d, seq_start:%d,%d\n", (int)bestoffset, (int)bestlength, (int)(ringbuffer_get(&buffer, offset)), (int)(ringbuffer_get(&buffer, offset - 1)));
		if(bestlength <= 2 || bestlength <= 3 && bestoffset >= 16)
		{
			fc = getc(in);
			ringbuffer_insert(&buffer, fc);

			if(fc != RLE_ESCAPE)
				putc(fc, out);
			else
			{
				putc(RLE_ESCAPE, out);
				putc(0, out);
			}
		}
		else
		{
			offsets[bestoffset]++;
			stream_to_buffer(&buffer, in, bestlength);
			putc(RLE_ESCAPE, out);
			if(bestoffset < 32 && bestlength <= 7 + 2)
			{
				bestlength--;
				putc((bestoffset << 3) | bestlength, out);
			}
			else
			{
				bestlength--;
				putc(((bestoffset >> 5) << 6) | ((bestlength >> 3) << 4), out);
				putc(((bestoffset & 0x1f) << 4) | (bestlength & 0x07), out);
			}
		}
	}
	//for(int z = 0; z < 128; z++) printf("o %d: %d\n", (int)z, (int)(offsets[z]));
}

int main(int argc, char *argv[])
{
	FILE *input, *output;
	char *output_filename;

	if(argc != 2)
	{
		fprintf(stderr, "ColecoVision LZK encoder\n");
		fprintf(stderr, "Usage: lzk_encoder <input file>\n");
		return(-1);
	}

	output_filename = malloc(strlen(argv[1]) + 5);
	if(!output_filename)
	{
		fprintf(stderr, "Out of memory.\n");
		return(-1);
	}
	strcpy(output_filename, argv[1]);
	strcpy(output_filename + strlen(argv[1]), ".lzk");
	input = fopen(argv[1], "rb");
	if(!output_filename)
	{
		fprintf(stderr, "Could not open input file %s\n", argv[1]);
		return(-1);
	}
	output = fopen(output_filename, "wb");
	if(!output_filename)
	{
		fprintf(stderr, "Could not open output file %s\n", output_filename);
		fclose(input);
		return(-1);
	}

	lzk_encode(input, output);

	fclose(input);
	fclose(output);
	return(0);
}

