// (c) 2006-2017 Philipp Klaus Krause philipp@colecovision.eu

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "rle_escape.h"

void rle_encode(FILE *in, FILE *out)
{
	int last;
	int current;
	int count;

	last = getc(in);
	if(last == EOF)
		return;
	count = 1;

	while((current = getc(in)) != EOF)
	{
		if(last == current && count > 0 && count <= 255)
		{
			count++;
		}
		else if((count == 1) && (last != RLE_ESCAPE))
		{
			putc(last, out);
			count = 1;
		}
		else if((count == 2) && (last != RLE_ESCAPE))
		{
			putc(last, out);
			putc(last, out);
			count = 1;
		}
		else
		{
			putc(RLE_ESCAPE, out);
			putc(count, out);
			putc(last, out);
			count = 1;
		}
		last = current;
	}

	if((count == 1) && (last != RLE_ESCAPE))
	{
		putc(last, out);
	}
	else if((count == 2) && (last != RLE_ESCAPE))
	{
		putc(last, out);
		putc(last, out);
	}
	else
	{
		putc(RLE_ESCAPE, out);
		putc(count, out);
		putc(last, out);
	}
}

int main(int argc, char *argv[])
{
	FILE *input, *output;
	char *output_filename;

	if(argc != 2)
	{
		fprintf(stderr, "ColecoVision RLE encoder\n");
		fprintf(stderr, "Usage: rle_encoder <input file>\n");
		return(-1);
	}

	output_filename = malloc(strlen(argv[1]) + 5);
	if(!output_filename)
	{
		fprintf(stderr, "Out of memory.\n");
		return(-1);
	}
	strcpy(output_filename, argv[1]);
	strcpy(output_filename + strlen(argv[1]), ".rle");
	input = fopen(argv[1], "rb");
	if(!input)
	{
		fprintf(stderr, "Could not open input file %s\n", argv[1]);
		return(-1);
	}
	output = fopen(output_filename, "wb");
	if(!output)
	{
		fprintf(stderr, "Could not open output file %s\n", output_filename);
		fclose(input);
		return(-1);
	}

	rle_encode(input, output);

	fclose(input);
	fclose(output);
	return(0);
}

